@@ -18,12 +18,19 @@ import asyncio
 import logging
 from copy import deepcopy
 from datetime import datetime, timedelta
+from random import random
 from urllib.parse import parse_qs, urlparse
 
 from asyncopenstackclient import AuthPassword, GlanceClient
 from simple_rest_client.exceptions import NotFoundError, ServerError
 
-from mrack.errors import ProvisioningError, ServerNotFoundError, ValidationError
+from mrack.errors import (
+    NotAuthenticatedError,
+    ProviderError,
+    ProvisioningError,
+    ServerNotFoundError,
+    ValidationError,
+)
 from mrack.host import STATUS_ACTIVE, STATUS_DELETED, STATUS_ERROR, STATUS_PROVISIONING
 from mrack.providers.provider import STRATEGY_RETRY, Provider
 from mrack.providers.utils.osapi import ExtraNovaClient, NeutronClient
@@ -37,6 +44,7 @@ logger = logging.getLogger(__name__)
 
 PROVISIONER_KEY = "openstack"
 SERVER_ERROR_RETRY = 5
+SERVER_ERROR_SLEEP = 10
 
 
 class OpenStackProvider(Provider):
@@ -66,7 +74,9 @@ class OpenStackProvider(Provider):
         self.timeout = 60  # minutes
         self.poll_sleep_initial = 15  # seconds
         self.poll_sleep = 7  # seconds
-        self.STATUS_MAP = {
+        self.poll_init_adj = 0  # set based on # of hosts to provisions
+        self.poll_adj = 0  # set based on # of hosts to provisions
+        self.status_map = {
             "ACTIVE": STATUS_ACTIVE,
             "BUILD": STATUS_PROVISIONING,
             "DELETED": STATUS_DELETED,
@@ -87,7 +97,14 @@ class OpenStackProvider(Provider):
         """
         # session expects that credentials will be set via env variables
         logger.info(f"{self.dsp_name}: Initializing provider")
-        self.session = AuthPassword()
+        try:
+            self.session = AuthPassword()
+        except TypeError as terr:
+            err = (
+                "OpenStack credentials not provided. Load OpenStack RC file with valid "
+                "credentials and try again. E.g.: $ source PROJECT-openrc.sh"
+            )
+            raise NotAuthenticatedError(err) from terr
         self.nova = ExtraNovaClient(session=self.session)
         self.glance = GlanceClient(session=self.session)
         self.neutron = NeutronClient(session=self.session)
@@ -99,8 +116,7 @@ class OpenStackProvider(Provider):
             self.neutron.init_api(self.api_timeout),
         )
         login_end = datetime.now()
-        login_duration = login_end - login_start
-        logger.info(f"{self.dsp_name}: Login duration {login_duration}")
+        logger.info(f"{self.dsp_name}: Login duration {login_end - login_start}")
 
         object_start = datetime.now()
         _, _, limits, _, _ = await asyncio.gather(
@@ -117,19 +133,19 @@ class OpenStackProvider(Provider):
             f"{self.dsp_name}: Environment objects load duration: {object_duration}"
         )
 
-    def set_flavors(self, flavors):
+    def _set_flavors(self, flavors):
         """Extend provider configuration with list of flavors."""
         for flavor in flavors:
             self.flavors[flavor["name"]] = flavor
             self.flavors_by_ref[flavor["id"]] = flavor
 
-    def set_images(self, images):
+    def _set_images(self, images):
         """Extend provider configuration with list of images."""
         for image in images:
             self.images[image["name"]] = image
             self.images_by_ref[image["id"]] = image
 
-    def set_networks(self, networks):
+    def _set_networks(self, networks):
         """Extend provider configuration with list of networks."""
         for network in networks:
             self.networks[network["name"]] = network
@@ -167,7 +183,7 @@ class OpenStackProvider(Provider):
         """Extend provider configuration by loading all flavors from OpenStack."""
         resp = await self.nova.flavors.list()
         flavors = resp["flavors"]
-        self.set_flavors(flavors)
+        self._set_flavors(flavors)
         return flavors
 
     async def load_images(self, image_names=None):
@@ -195,12 +211,12 @@ class OpenStackProvider(Provider):
             query = p_result.query
             next_params = parse_qs(query)
             for key, val in next_params.items():
-                if type(val) == list and len(val):
+                if isinstance(val, list) and val:
                     next_params[key] = val[0]
             response = await self.glance.images.list(**next_params)
             images.extend(response["images"])
 
-        self.set_images(images)
+        self._set_images(images)
 
         return images
 
@@ -208,7 +224,7 @@ class OpenStackProvider(Provider):
         """Extend provider configuration by loading all networks from OpenStack."""
         resp = await self.neutron.network.list()
         networks = resp["networks"]
-        self.set_networks(networks)
+        self._set_networks(networks)
         return networks
 
     async def load_ip_availabilities(self):
@@ -252,7 +268,7 @@ class OpenStackProvider(Provider):
         network_specs = req.get("networks", [])
         network_specs = deepcopy(network_specs)
         networks = []
-        if type(network_specs) != list:
+        if not isinstance(network_specs, list):
             network_specs = []
         for network_spec in network_specs:
             uuid = network_spec.get("uuid")
@@ -279,6 +295,23 @@ class OpenStackProvider(Provider):
 
         return True
 
+    def _set_poll_sleep_times(self, reqs):
+        """
+        Compute polling sleep times based on number of hosts.
+
+        So that we don't create unnecessary load on server while still checking returns
+        (initial_sleep, sleep)
+        """
+        count = len(reqs)
+
+        # initial poll is the biggest performance saver it should be around
+        # time when more than half of host is in ACTIVE state
+        self.poll_init_adj = 0.65 * count
+
+        # poll time should ask often enough, to not create unnecessary delays
+        # while not that many to not load the server much
+        self.poll_adj = 0.22 * count
+
     async def prepare_provisioning(self, reqs):
         """
         Prepare provisioning.
@@ -286,14 +319,17 @@ class OpenStackProvider(Provider):
         Load missing images if they are not in provisioning-config.yaml
         """
         prepare_images = list(
-            set([req["image"] for req in reqs if req["image"] not in self.images])
+            {req["image"] for req in reqs if req["image"] not in self.images}
         )
+
         if prepare_images:
             im_list = ", ".join(prepare_images)
             logger.debug(f"{self.dsp_name}: Loading image info for: '{im_list}'")
             await self.load_images(list(prepare_images))
             logger.debug(f"{self.dsp_name}: Loading images info done.")
 
+        self._set_poll_sleep_times(reqs)
+
     async def validate_hosts(self, reqs):
         """Validate that all hosts requirements contains existing required objects."""
         for req in reqs:
@@ -303,13 +339,23 @@ class OpenStackProvider(Provider):
         """Get vCPU and memory requirements for host requirement."""
         flavor_spec = req.get("flavor")
         flavor_ref = req.get("flavorRef")
+        flavor = None
         if flavor_ref:
             flavor = self.get_flavor(ref=flavor_ref)
         if flavor_spec:
             flavor = self.get_flavor(flavor_spec, flavor_spec)
-        return {"ram": flavor["ram"], "vcpus": flavor["vcpus"]}
 
-    async def can_provision(self, reqs):
+        try:
+            res = {"ram": flavor["ram"], "vcpus": flavor["vcpus"]}
+        except TypeError as flavor_none:
+            # func does not load flavor so None is used as result
+            raise ValidationError(
+                f"Could not load the flavor for requirement: {req}"
+            ) from flavor_none
+
+        return res
+
+    async def can_provision(self, reqs):  # pylint: disable=arguments-differ
         """Check that all host can be provisioned.
 
         Checks:
@@ -376,7 +422,20 @@ class OpenStackProvider(Provider):
         if specs.get("network"):
             del specs["network"]
 
-        response = await self.nova.servers.create(server=specs)
+        error_attempts = 0
+        while True:
+            try:
+                response = await self.nova.servers.create(server=specs)
+            except ServerError as exc:
+                logger.debug(exc)
+                error_attempts += 1
+                if error_attempts > SERVER_ERROR_RETRY:
+                    raise ProvisioningError(
+                        f"{self.dsp_name}: Fail to create server", specs
+                    ) from exc
+                await asyncio.sleep(SERVER_ERROR_SLEEP)
+            else:
+                break
         return response.get("server")
 
     async def delete_server(self, uuid):
@@ -384,17 +443,24 @@ class OpenStackProvider(Provider):
 
         Doesn't wait for the deletion to happen.
         """
-        try:
-            await self.nova.servers.force_delete(uuid)
-        except NotFoundError:
-            logger.warning(
-                f"{self.dsp_name}: Server '{uuid}' not found, probably already deleted"
-            )
-            pass
+        error_attempts = 0
+        while True:
+            try:
+                await self.nova.servers.force_delete(uuid)
+            except ServerError as exc:
+                logger.debug(exc)
+                error_attempts += 1
+                if error_attempts > SERVER_ERROR_RETRY:
+                    raise ProviderError(uuid) from exc
+                await asyncio.sleep(SERVER_ERROR_SLEEP)
+            except NotFoundError:
+                logger.warning(
+                    f"{self.dsp_name}: Server '{uuid}' no found, probably already "
+                    "deleted"
+                )
+                break
 
-    async def wait_till_provisioned(
-        self, instance, timeout=None, poll_sleep=None, poll_sleep_initial=None
-    ):
+    async def wait_till_provisioned(self, resource):
         """
         Wait till server is provisioned.
 
@@ -410,37 +476,42 @@ class OpenStackProvider(Provider):
 
         Return information about provisioned server.
         """
-        uuid = instance.get("id")
-        if not poll_sleep_initial:
-            poll_sleep_initial = self.poll_sleep_initial
-        if not poll_sleep:
-            poll_sleep = self.poll_sleep
-        if not timeout:
-            timeout = self.timeout
+        uuid = resource.get("id")
+
+        poll_sleep_initial = self.poll_sleep_initial + self.poll_init_adj
+        poll_sleep_initial = (
+            poll_sleep_initial / 2 + poll_sleep_initial * random() * 1.5
+        )
+        poll_sleep = self.poll_sleep + self.poll_adj
+        timeout = self.timeout
 
         start = datetime.now()
         timeout_time = start + timedelta(minutes=timeout)
-        done_states = ["ACTIVE", "ERROR"]
 
         # do not check the state immediately, it will take some time
+        logger.debug(f"{uuid}: sleeping for {poll_sleep_initial} seconds")
         await asyncio.sleep(poll_sleep_initial)
 
-        error_attempts = 0
+        resp = {}
         logger.debug(f"Waiting for: {uuid}")
+        error_attempts = 0
         while datetime.now() < timeout_time:
             try:
                 resp = await self.nova.servers.get(uuid)
-                server = resp["server"]
-                if server["status"] in done_states:
-                    break
-            except NotFoundError:
-                raise ServerNotFoundError(uuid)
-            except ServerError as e:
-                logger.debug(e)
+            except NotFoundError as nf_err:
+                raise ServerNotFoundError(uuid) from nf_err
+            except ServerError as err:
+                logger.debug(f"{self.dsp_name}: {err}")
                 error_attempts += 1
                 if error_attempts > SERVER_ERROR_RETRY:
-                    raise ProvisioningError(uuid)
+                    raise ProvisioningError(uuid) from err
 
+            server = resp["server"]
+            if server["status"] in ["ACTIVE", "ERROR"]:
+                break
+
+            poll_sleep += 0.5  # increase delays to check the longer it takes
+            logger.debug(f"{uuid}: sleeping for {poll_sleep} seconds")
             await asyncio.sleep(poll_sleep)
 
         done_time = datetime.now()
@@ -465,14 +536,8 @@ class OpenStackProvider(Provider):
         return True
 
     def prov_result_to_host_data(self, prov_result):
-        """Get needed host infromation from openstack provisioning result."""
-        result = {
-            "id": None,
-            "name": None,
-            "addresses": None,
-            "status": None,
-            "fault": None,
-        }
+        """Get needed host information from openstack provisioning result."""
+        result = {}
 
         result["id"] = prov_result.get("id")
         result["name"] = prov_result.get("name")
